/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-2008 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.descriptor.original;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URL;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * This Class defines the JBI Descriptor builder for the kernel. <li>It can
 * validate JBI descriptor in XML form. <li>It can build class bindings from a
 * XML form. <li>It can build the XML binding from classes form.
 * <p>
 * Note that this Builder is based on the original JBI schema; it doesn't
 * control the JBI extension parts.
 * </p>
 * <p>
 * This class is thread-safe.
 * </p>
 * 
 * @author Roland NAUDIN - EBM WebSourcing
 * @since PEtALS 2.2
 */
final public class JBIDescriptorBuilder {

    /**
     * JBI schema name
     */
    public static final String XSD_JBI = "jbi.xsd";

    /**
     * Standard location to the JBI Descriptor resource.
     */
    public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF" + File.separator + "jbi.xml";

    /**
     * The JAXB unique unmarshaller.
     */
    private static Unmarshaller unmarshaller = null;

    private static Object unmarshallerSync;

    /**
     * The exception raised during JAXB unique unmarshaller creation.
     */
    private static JBIDescriptorException unmarshallerCreationEx = null;

    /**
     * The JAXB unique marshaller.
     */
    private static Marshaller marshaller = null;

    private static Object marshallerSync;

    /**
     * The exception raised during JAXB unique marshaller creation.
     */
    private static JBIDescriptorException marshallerCreationEx = null;

    /*
     * Private object initializations
     */
    static {
        final SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        try {
            // The JBI schema resource is in the Jar where the class is
            // loaded
            final URL schemaUrl = JBIDescriptorBuilder.class.getResource("/" + XSD_JBI);
            final Schema schema = factory.newSchema(new StreamSource[] { new StreamSource(schemaUrl
                    .openStream()) });

            // Unmarshaller initialization
            try {
                final JAXBContext jaxbContext = JAXBContext.newInstance(new Class[] { Jbi.class });
                unmarshaller = jaxbContext.createUnmarshaller();
                unmarshaller.setSchema(schema);
            } catch (JAXBException e) {
                unmarshallerCreationEx = new JBIDescriptorException(
                        "Failed to create the JAXB unmarshaller", e);
            }
            unmarshallerSync = new Object();

            // Marshaller initialization
            try {
                final JAXBContext jaxbContext = JAXBContext.newInstance(new Class[] { Jbi.class });
                marshaller = jaxbContext.createMarshaller();
                marshaller.setSchema(schema);
            } catch (JAXBException e) {
                marshallerCreationEx = new JBIDescriptorException(
                        "Failed to create the JAXB marshaller", e);
            }
            marshallerSync = new Object();

        } catch (final IOException e) {
            final JBIDescriptorException schemaCreationEx = new JBIDescriptorException(
                    "Failed to get resource '" + XSD_JBI + "'from the current class-path", e);
            unmarshallerCreationEx = new JBIDescriptorException(
                    "Failed to create the JAXB unmarshaller", schemaCreationEx);

            marshallerCreationEx = new JBIDescriptorException(
                    "Failed to create the JAXB marshaller", schemaCreationEx);
        } catch (final SAXException e) {
            final JBIDescriptorException schemaCreationEx = new JBIDescriptorException(
                    "Failed to parse resource '" + XSD_JBI + "'", e);
            unmarshallerCreationEx = new JBIDescriptorException(
                    "Failed to create the JAXB unmarshaller", schemaCreationEx);

            marshallerCreationEx = new JBIDescriptorException(
                    "Failed to create the JAXB marshaller", schemaCreationEx);
        }
    }

    /**
     * The class is static, so no constructor available.
     */
    private JBIDescriptorBuilder() {
        // NOP
    }

    /**
     * Validate the given JBI descriptor against the JBI schema.
     * 
     * @param jbiDescriptorStream
     * @throws JBIException
     *      The validation has failed
     * @throws JBIDescriptorException
     *      Failed to create the validator
     */
    public static void validateJBIDescriptor(final InputStream jbiDescriptorStream)
            throws JBIDescriptorException {

        // TODO : is the validator thread safe?
        final Validator validator = getUnMarshaller().getSchema().newValidator();

        try {
            validator.validate(new StreamSource(jbiDescriptorStream));
        } catch (final SAXException e) {
            throw new JBIDescriptorException(
                    "Failed to validate JBI descriptor against JBI schema", e);
        } catch (final IOException e) {
            throw new JBIDescriptorException(
                    "Failed to validate JBI descriptor against JBI schema", e);
        }
    }

    /**
     * <p>
     * Build the Java bindings from the JBI descriptor in XML form. The
     * validation against the JBI schema is processed during the transformation.
     * </p>
     * <p>
     * <b>Note</b>: The input stream is closed because of the XML parser.
     * </p>
     * 
     * @param jbiDescriptorStream
     *            The JBI descriptor in XML form as {@link InputStream}
     * @return The root object (instance of {@link Jbi}) of the JBI descriptor
     * @throws JBIDescriptorException
     *             The exception raised during the unmarshaller creation or the
     *             exception raised during the build of the java bindings.
     */
    public static Jbi buildJavaJBIDescriptor(final InputStream jbiDescriptorStream)
            throws JBIDescriptorException {

        try {
            final JAXBElement<Jbi> jbiBinding;

            // The default Xerces unmarshaller is not thread safe
            synchronized (unmarshallerSync) {
                jbiBinding = getUnMarshaller().unmarshal(new StreamSource(jbiDescriptorStream),
                        Jbi.class);
            }

            return jbiBinding.getValue();

        } catch (JAXBException e) {
            throw new JBIDescriptorException(
                    "Failed to build Java bindings from JBI descriptor XML document", e);
        }
    }

    /**
     * Build the XML nodes from the JBI descriptor in Java classes form.
     * 
     * @param jbiDescriptorClass
     *      The JBI Descriptor root class
     * @param jbiDescriptorNode
     *      The XML Node to fill with the JBI descriptor XML nodes
     * @throws JBIDescriptorException
     *      The exception raised during the unmarshaller creation or the
     *      exception raised during the build of the XML nodes.
     */
    public static void buildXmlJBIdescriptor(final Jbi jbiDescriptorClass,
            final Node jbiDescriptorNode) throws JBIDescriptorException {

        try {
            // The default Xerces marshaller is not thread safe
            synchronized (marshallerSync) {
                getMarshaller().marshal(jbiDescriptorClass, jbiDescriptorNode);
            }
        } catch (JAXBException e) {
            throw new JBIDescriptorException(
                    "Failed to build XML binding from JBI descriptor Java classes", e);
        }
    }

    /**
     * Build the XML String from the JBI descriptor in Java classes form.
     * 
     * @param jbiDescriptorClass
     *      The JBI Descriptor root class
     * @return The String to fill with the JBI descriptor XML
     * @throws JBIDescriptorException
     *      The exception raised during the marshaller creation or the exception
     *      raised during the build of the XML string.
     */
    public static String buildXmlStringJBIdescriptor(final Jbi jbiDescriptorClass)
            throws JBIDescriptorException {

        try {
            final StringWriter stringWriter = new StringWriter();
            // The default Xerces marshaller is not thread safe
            synchronized (marshallerSync) {
                getMarshaller().marshal(jbiDescriptorClass, stringWriter);
            }
            
            return stringWriter.toString();
        } catch (JAXBException e) {
            throw new JBIDescriptorException(
                    "Failed to build XML binding from JBI descriptor Java classes", e);
        }
    }

    /**
     * Get the unmarshaller instance.
     * 
     * @return the unmarshaller instance
     * @throws JBIDescriptorException
     *      The exception raised during the unmarshaller creation.
     */
    private static Unmarshaller getUnMarshaller() throws JBIDescriptorException {

        if (unmarshallerCreationEx != null) {
            throw unmarshallerCreationEx;
        }

        return unmarshaller;
    }

    /**
     * Get the marshaller instance.
     * 
     * @return the marshaller instance
     * @throws JBIDescriptorException
     *      The exception raised during the marshaller creation.
     */
    private static Marshaller getMarshaller() throws JBIDescriptorException {

        if (marshallerCreationEx != null) {
            throw marshallerCreationEx;
        }

        return marshaller;
    }
}
